# -*- coding: utf-8 -*-
"""
yadt Python wrapper for YaDT decisionb tree builder
version 1.0

@author: Salvatore Ruggieri (c) 2018-2020
"""

# imports
import os
import csv
import subprocess
import numpy as np
import pandas as pd
from collections import OrderedDict
from sklearn.base import BaseEstimator, ClassifierMixin
import shutil

# set working directory
def setup(working_dir=os.path.abspath(os.path.dirname(__name__))):
    global wdir
    wdir = os.path.join(working_dir, '__yadtcache__')   
    if not os.path.exists(wdir):
        os.mkdir(wdir)

# default setup
setup()

# cache cleaning
def clean():
    global wdir
    if os.path.exists(wdir):
        shutil.rmtree(wdir)

# extract metadata for YaDT from DataFrame
def meta_data(df, ovr_types={}):
    res = OrderedDict()
    for col in df.columns:
        dtype = df[col].dtype
        if dtype == np.float64:
            col_type = 'double'
            feat_type = 'continuous'
        else:
            kind = dtype.kind
            if kind == 'f':
                col_type = 'float'
                feat_type = 'continuous'
            elif kind == 'i':
                col_type = 'integer'
                feat_type = 'continuous'
            elif kind == 'u':
                col_type = 'uinteger'
                feat_type = 'continuous'
            else: # category
                col_type = 'string'        
                feat_type = 'discrete'
        if col in ovr_types:
            col_type, feat_type = ovr_types[col]
        res[col] = (col_type, feat_type)
    return res

# export DataFrame in YaDT format
def to_yadt(X, metadata, y=None, filenames=None, filedata=None, sep=',', exclude_names=False):
    if not isinstance(metadata, OrderedDict):
        raise "metadata not an OrderedDict"
    meta = metadata
    if y is not None and y.name in meta:
        del meta[y.name]
    if isinstance(X, np.ndarray):
        # assume columns are in order with meta
        alldata = pd.DataFrame(X, columns=meta.keys())
    else:
        alldata = X[meta]
    if y is not None:
        meta[y.name] = ('string', 'class')
        #print(type(y))
        alldata[y.name] = y.values
    # output meta-data
    if filenames is not None:
        with open(filenames, 'w') as names_file:
            for col in meta.keys():
                print("{}{}{}{}{}".format(col, sep, meta[col][0], sep, meta[col][1]), file=names_file)
    if filedata is not None:
        alldata.to_csv(filedata, na_rep='?', 
                   quoting=csv.QUOTE_NONE, sep=sep, index=None,
                   float_format='%g', header=False)
 
class YaDTClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, metadata, sep=',', options=""):
        self.metadata = metadata
        self.sep = sep
        self.options = options
        self.target = None
        self.tree = None
        
    def get_data_md(self, X, weights=None, exclude_class=False):
        md = self.metadata
        alldata = X
        if weights is not None:
            md[weights.name] = ('double', 'weights')
            if isinstance(X, pd.DataFrame):
                alldata[weights.name] = weights
            elif isinstance(X, np.ndarray):
                wcol = np.reshape(weights, (weights.shape[0], 1))
                alldata = np.append((alldata, wcol), axis=1)
        if exclude_class and self.target in md:
            del md[self.target]
        return alldata, md
 
    def fit(self, X, y, weights=None, verbose=False):
        global wdir
        alldata, md = self.get_data_md(X, weights)
        self.target = y.name if y is not None else None
        file_base = str(id(X))
        namesfile = os.path.join(wdir,file_base+'.names')
        datafile = os.path.join(wdir,file_base+'.data.gz')
        to_yadt(alldata, md, y=y, filenames=namesfile, filedata=datafile, sep=self.sep)
        treefile = os.path.join(wdir,file_base+'.tree')
        cmdfile = os.path.join(wdir,'..', 'dTcmd')
        cmd = cmdfile + ' -fm {} -fd {} -sep "{}" -tb {} {}'.format(
                os.path.join(wdir,namesfile), 
                os.path.join(wdir,datafile), self.sep, 
                treefile, self.options)
        if verbose:
            print(cmd)
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True)
        if verbose:
            print(output)
        with open(treefile, mode='rb') as file:
            self.tree = file.read()
        os.remove(treefile)
        os.remove(datafile)
        os.remove(namesfile)
        return self

    def load_tree(self, filename, verbose=False):
        global wdir
        treefile = os.path.join(wdir,filename)
        with open(treefile, mode='rb') as file:
            self.tree = file.read()
        return self

    def predict(self, X, weights=None, verbose=False):
        alldata, md = self.get_data_md(X, weights, exclude_class=True)
        file_base = str(id(self))
        datafile = os.path.join(wdir,file_base+'.score.data.gz')
        to_yadt(alldata, md, filedata=datafile, sep=self.sep)
        treefile = os.path.join(wdir,file_base+'.tree')
        scorefile = os.path.join(wdir,file_base+'.score')
        with open(treefile, mode='wb') as file:
            file.write(self.tree)
        cmdfile = os.path.join(wdir,'..', 'dTcmd')
        cmd = cmdfile + ' -bt {} -fs {} -sep "{}" -s {}'.format(
                treefile, datafile, self.sep, scorefile)
        if verbose:
            print(cmd)
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True)
        if verbose:
            print(output)
        res = pd.read_csv(scorefile, sep=self.sep, header=None)
        os.remove(treefile)
        os.remove(datafile)
        os.remove(scorefile)
        return res[0].values

    def predict_proba(self, X, weights=None, verbose=False):
        alldata, md = self.get_data_md(X, weights, exclude_class=True)
        file_base = str(id(self))
        datafile = os.path.join(wdir,file_base+'.score.data.gz')
        to_yadt(alldata, md, filedata=datafile, sep=self.sep)
        treefile = os.path.join(wdir,file_base+'.tree')
        scorefile = os.path.join(wdir,file_base+'.score')
        with open(treefile, mode='wb') as file:
            file.write(self.tree)
        cmdfile = os.path.join(wdir,'..', 'dTcmd')
        cmd = cmdfile + ' -bt {} -fs {} -sep "{}" -s {} -sp'.format(
                treefile, datafile, self.sep, scorefile)
        if verbose:
            print(cmd)
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True)
        if verbose:
            print(output)
        res = pd.read_csv(scorefile, sep=self.sep, header=None)
        os.remove(treefile)
        os.remove(datafile)
        os.remove(scorefile)
        return res.iloc[:,2:].values

    def get_tree(self, form="string", verbose=False):
        file_base = str(id(self))
        treefile = os.path.join(wdir,file_base+'.tree')
        formatfile = os.path.join(wdir,file_base+'.txt')
        with open(treefile, mode='wb') as file:
            file.write(self.tree)
        cmdfile = os.path.join(wdir,'..', 'dTcmd')
        fmt = {"string":"-t", "dot":"-d", "xml":"-x"}
        if form not in fmt:
            form = "string"
        cmd = cmdfile + ' -bt {} {} {}'.format(treefile, fmt[form], formatfile)
        if verbose:
            print(cmd)
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True)
        if verbose:
            print(output)
        with open(formatfile, mode='rt', encoding="utf-8") as file:
            res = file.read()
        os.remove(treefile)
        os.remove(formatfile)
        return res
